__all__ = ["Client"]

import json
import time
from typing import Any, Literal

from unicex._base import BaseClient
from unicex.exceptions import NotAuthorized
from unicex.types import RequestMethod
from unicex.utils import dict_to_query_string, filter_params, generate_hmac_sha256_signature


class Client(BaseClient):
    """Клиент для работы с Bybit API."""

    _BASE_URL: str = "https://api.bybit.com"
    """Базовый URL для REST API Bybit."""

    _RECV_WINDOW: str = "5000"
    """Стандартный интервал времени для получения ответа от сервера."""

    def _get_headers(self, timestamp: str, signature: str | None = None) -> dict:
        """Возвращает заголовки для запросов к Bybit API.

        Параметры:
            timestamp (str): Временная метка запроса в миллисекундах.
            signature (str | None): Подпись запроса, если запрос авторизированый.
        """
        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        if signature:
            headers["X-BAPI-API-KEY"] = self._api_key  # type: ignore
            headers["X-BAPI-SIGN-TYPE"] = "2"
            headers["X-BAPI-SIGN"] = signature
            headers["X-BAPI-RECV-WINDOW"] = self._RECV_WINDOW
            headers["X-BAPI-TIMESTAMP"] = timestamp
        return headers

    def _generate_signature(self, timestamp: str, payload: dict, method: RequestMethod) -> str:
        """Генерация подписи.

        Источник: https://github.com/bybit-exchange/api-usage-examples/blob/master/V5_demo/api_demo/Encryption_HMAC.py
        """
        # Проверяем наличие апи ключей для подписи запроса
        if not self.is_authorized():
            raise NotAuthorized("Api key and api secret is required to private endpoints")

        if method == "POST":
            # timestamp+api_key+recv_window+jsonBodyString
            dumped_payload = json.dumps(payload)
            prepared_query_string = timestamp + self._api_key + self._RECV_WINDOW + dumped_payload  # type: ignore[attrDefined]
            return generate_hmac_sha256_signature(self._api_secret, prepared_query_string)  # type: ignore[attrDefined]
        else:
            # timestamp+api_key+recv_window+queryString
            query_string = dict_to_query_string(payload)
            prepared_query_string = timestamp + self._api_key + self._RECV_WINDOW + query_string  # type: ignore[attrDefined]
            return generate_hmac_sha256_signature(self._api_secret, prepared_query_string)  # type: ignore[attrDefined]

    async def _make_request(
        self,
        method: RequestMethod,
        endpoint: str,
        *,
        params: dict[str, Any] | None = None,
        signed: bool = False,
    ) -> Any:
        """Выполняет HTTP-запрос к эндпоинтам Bybit API с поддержкой подписи.

        Если signed=True, формируется подпись для приватных endpoint'ов.
        Если signed=False, запрос отправляется как обычный публичный, через
        базовый _make_request без обработки подписи.

        Параметры:
            method (str): HTTP метод запроса ("GET", "POST", "DELETE" и т.д.).
            endpoint (str): URL эндпоинта Bybit API.
            params (dict | None): Параметры запроса. Передаются в body, если запрос типа "POST", иначе в query_params
            signed (bool): Нужно ли подписывать запрос.

        Возвращает:
            dict: Ответ в формате JSON.
        """
        # Составляем URL для запроса
        url = self._BASE_URL + endpoint

        # Фильтруем параметры от None значений
        params = filter_params(params) if params else {}

        # Генерируем временную метку
        timestamp = str(int(time.time() * 1000))

        # Проверяем нужно ли подписывать запрос
        if not signed:
            headers = self._get_headers(timestamp)
            return await super()._make_request(
                method=method,
                url=url,
                headers=headers,
                params=params,
            )

        # Формируем payload
        payload = params

        # Генерируем строку для подписи
        signature = self._generate_signature(timestamp, payload, method)

        # Генерируем заголовки (вкл. в себя подпись и апи ключ)
        headers = self._get_headers(timestamp, signature)

        if method == "POST":  # Отправляем параметры в тело запроса
            return await super()._make_request(
                method=method,
                url=url,
                data=payload,
                headers=headers,
            )
        else:  # Иначе параметры добавляем к query string
            return await super()._make_request(
                method=method,
                url=url,
                params=params,
                headers=headers,
            )

    async def request(
        self, method: RequestMethod, endpoint: str, params: dict, signed: bool
    ) -> dict:
        """Специальный метод для выполнения запросов на эндпоинты, которые не обернуты в клиенте.

        Параметры:
            method (RequestMethod): Метод запроса (GET, POST, PUT, DELETE).
            endpoint (str): URL эндпоинта.
            params (dict): Параметры запроса.
            signed (bool): Флаг, указывающий, требуется ли подпись запроса.

        Возвращает:
            `dict`: Ответ в формате JSON.
        """
        return await self._make_request(
            method=method, endpoint=endpoint, params=params, signed=signed
        )

    # topic: market

    async def ping(self) -> dict:
        """Проверка подключения к REST API.

        https://bybit-exchange.github.io/docs/v5/market/time
        """
        return await self._make_request("GET", "/v5/market/time")

    async def klines(
        self,
        symbol: str,
        interval: str,
        category: Literal["spot", "linear", "inverse"],
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Исторические свечи.

        https://bybit-exchange.github.io/docs/v5/market/kline
        """
        params = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }

        return await self._make_request("GET", "/v5/market/kline", params=params)

    async def mark_price_klines(
        self,
        symbol: str,
        interval: str,
        category: Literal["linear", "inverse"] | None = None,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Исторические свечи цены маркировки (mark price).

        https://bybit-exchange.github.io/docs/v5/market/mark-kline
        """
        params = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }

        return await self._make_request("GET", "/v5/market/mark-price-kline", params=params)

    async def index_price_klines(
        self,
        symbol: str,
        interval: str,
        category: Literal["linear", "inverse"] | None = None,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Исторические свечи индекса (index price).

        https://bybit-exchange.github.io/docs/v5/market/index-kline
        """
        params = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }

        return await self._make_request("GET", "/v5/market/index-price-kline", params=params)

    async def premium_index_price_klines(
        self,
        symbol: str,
        interval: str,
        category: Literal["linear"] | None = None,
        start: int | None = None,
        end: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Исторические свечи премиального индекса (premium index).

        https://bybit-exchange.github.io/docs/v5/market/premium-index-kline
        """
        params = {
            "category": category,
            "symbol": symbol,
            "interval": interval,
            "start": start,
            "end": end,
            "limit": limit,
        }

        return await self._make_request(
            "GET", "/v5/market/premium-index-price-kline", params=params
        )

    async def instruments_info(
        self,
        category: Literal["spot", "linear", "inverse", "option"],
        symbol: str | None = None,
        status: str | None = None,
        base_coin: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Информация об инструментах.

        https://bybit-exchange.github.io/docs/v5/market/instrument
        """
        params = {
            "category": category,
            "symbol": symbol,
            "status": status,
            "baseCoin": base_coin,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request("GET", "/v5/market/instruments-info", params=params)

    async def orderbook(
        self,
        category: Literal["spot", "linear", "inverse", "option"],
        symbol: str,
        limit: int | None = None,
    ) -> dict:
        """Книга ордеров (глубина рынка).

        https://bybit-exchange.github.io/docs/v5/market/orderbook
        """
        params = {
            "category": category,
            "symbol": symbol,
            "limit": limit,
        }

        return await self._make_request("GET", "/v5/market/orderbook", params=params)

    async def rpi_orderbook(
        self,
        symbol: str,
        limit: int,
        category: Literal["spot", "linear", "inverse"] | None = None,
    ) -> dict:
        """Книга ордеров RPI.

        https://bybit-exchange.github.io/docs/v5/market/rpi-orderbook
        """
        params = {
            "category": category,
            "symbol": symbol,
            "limit": limit,
        }

        return await self._make_request("GET", "/v5/market/rpi_orderbook", params=params)

    async def tickers(
        self,
        category: Literal["spot", "linear", "inverse", "option"],
        symbol: str | None = None,
        base_coin: str | None = None,
        exp_date: str | None = None,
    ) -> dict:
        """Тикеры (снимок цен и объёмов).

        https://bybit-exchange.github.io/docs/v5/market/tickers
        """
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "expDate": exp_date,
        }

        return await self._make_request("GET", "/v5/market/tickers", params=params)

    async def funding_rate_history(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """История ставок финансирования.

        https://bybit-exchange.github.io/docs/v5/market/history-fund-rate
        """
        params = {
            "category": category,
            "symbol": symbol,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
        }

        return await self._make_request("GET", "/v5/market/funding/history", params=params)

    async def recent_trades(
        self,
        category: Literal["spot", "linear", "inverse", "option"],
        symbol: str | None = None,
        base_coin: str | None = None,
        option_type: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Недавние публичные сделки.

        https://bybit-exchange.github.io/docs/v5/market/recent-trade
        """
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "optionType": option_type,
            "limit": limit,
        }

        return await self._make_request("GET", "/v5/market/recent-trade", params=params)

    async def open_interest(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        interval_time: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Открытый интерес по символу.

        https://bybit-exchange.github.io/docs/v5/market/open-interest
        """
        params = {
            "category": category,
            "symbol": symbol,
            "intervalTime": interval_time,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request("GET", "/v5/market/open-interest", params=params)

    async def historical_volatility(
        self,
        category: Literal["option"],
        base_coin: str | None = None,
        quote_coin: str | None = None,
        period: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> dict:
        """Историческая волатильность (опционы).

        https://bybit-exchange.github.io/docs/v5/market/iv
        """
        params = {
            "category": category,
            "baseCoin": base_coin,
            "quoteCoin": quote_coin,
            "period": period,
            "startTime": start_time,
            "endTime": end_time,
        }

        return await self._make_request("GET", "/v5/market/historical-volatility", params=params)

    async def insurance_pool(
        self,
        coin: str | None = None,
    ) -> dict:
        """Данные страхового пула.

        https://bybit-exchange.github.io/docs/v5/market/insurance
        """
        params = {
            "coin": coin,
        }

        return await self._make_request("GET", "/v5/market/insurance", params=params)

    async def risk_limit(
        self,
        category: Literal["linear", "inverse"],
        symbol: str | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Параметры лимита риска.

        https://bybit-exchange.github.io/docs/v5/market/risk-limit
        """
        params = {
            "category": category,
            "symbol": symbol,
            "cursor": cursor,
        }

        return await self._make_request("GET", "/v5/market/risk-limit", params=params)

    async def delivery_price(
        self,
        category: Literal["linear", "inverse", "option"],
        symbol: str | None = None,
        base_coin: str | None = None,
        settle_coin: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Цена поставки (delivery price).

        https://bybit-exchange.github.io/docs/v5/market/delivery-price
        """
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request("GET", "/v5/market/delivery-price", params=params)

    async def new_delivery_price(
        self,
        category: Literal["option"],
        base_coin: str,
        settle_coin: str | None = None,
    ) -> dict:
        """Исторические цены поставки (опционы).

        https://bybit-exchange.github.io/docs/v5/market/new-delivery-price
        """
        params = {
            "category": category,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
        }

        return await self._make_request("GET", "/v5/market/new-delivery-price", params=params)

    async def long_short_ratio(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        period: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Лонг/шорт соотношение.

        https://bybit-exchange.github.io/docs/v5/market/long-short-ratio
        """
        params = {
            "category": category,
            "symbol": symbol,
            "period": period,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request("GET", "/v5/market/account-ratio", params=params)

    async def index_price_components(
        self,
        index_name: str,
    ) -> dict:
        """Компоненты индексной цены.

        https://bybit-exchange.github.io/docs/v5/market/index-components
        """
        params = {
            "indexName": index_name,
        }

        return await self._make_request("GET", "/v5/market/index-price-components", params=params)

    async def order_price_limit(
        self,
        symbol: str,
        category: Literal["spot", "linear", "inverse"] | None = None,
    ) -> dict:
        """Лимиты цен для ордеров.

        https://bybit-exchange.github.io/docs/v5/market/order-price-limit
        """
        params = {
            "category": category,
            "symbol": symbol,
        }

        return await self._make_request("GET", "/v5/market/price-limit", params=params)

    async def adl_alert(
        self,
        symbol: str | None = None,
    ) -> dict:
        """Уведомления ADL и информация страхового пула.

        https://bybit-exchange.github.io/docs/v5/market/adl-alert
        """
        params = {
            "symbol": symbol,
        }

        return await self._make_request("GET", "/v5/market/adlAlert", params=params)

    async def fee_group_info(
        self,
        product_type: Literal["contract"],
        group_id: str | None = None,
    ) -> dict:
        """Структура групп комиссий и ставки.

        https://bybit-exchange.github.io/docs/v5/market/fee-group-info
        """
        params = {
            "productType": product_type,
            "groupId": group_id,
        }

        return await self._make_request("GET", "/v5/market/fee-group-info", params=params)

    # topic: trade

    async def create_order(
        self,
        category: Literal["linear", "inverse", "spot", "option"],
        symbol: str,
        side: Literal["Buy", "Sell"],
        order_type: Literal["Market", "Limit"],
        qty: str,
        is_leverage: int | None = None,
        market_unit: str | None = None,
        slippage_tolerance_type: str | None = None,
        slippage_tolerance: str | None = None,
        price: str | None = None,
        trigger_direction: int | None = None,
        order_filter: str | None = None,
        trigger_price: str | None = None,
        trigger_by: str | None = None,
        order_iv: str | None = None,
        time_in_force: str | None = None,
        position_idx: int | None = None,
        order_link_id: str | None = None,
        take_profit: str | None = None,
        stop_loss: str | None = None,
        tp_trigger_by: str | None = None,
        sl_trigger_by: str | None = None,
        reduce_only: bool | None = None,
        close_on_trigger: bool | None = None,
        smp_type: str | None = None,
        mmp: bool | None = None,
        tpsl_mode: Literal["Full", "Partial"] | None = None,
        tp_limit_price: str | None = None,
        sl_limit_price: str | None = None,
        tp_order_type: str | None = None,
        sl_order_type: str | None = None,
    ) -> dict:
        """Создание ордера.

        https://bybit-exchange.github.io/docs/v5/order/create-order
        """
        params = {
            "category": category,
            "symbol": symbol,
            "isLeverage": is_leverage,
            "side": side,
            "orderType": order_type,
            "qty": qty,
            "marketUnit": market_unit,
            "slippageToleranceType": slippage_tolerance_type,
            "slippageTolerance": slippage_tolerance,
            "price": price,
            "triggerDirection": trigger_direction,
            "orderFilter": order_filter,
            "triggerPrice": trigger_price,
            "triggerBy": trigger_by,
            "orderIv": order_iv,
            "timeInForce": time_in_force,
            "positionIdx": position_idx,
            "orderLinkId": order_link_id,
            "takeProfit": take_profit,
            "stopLoss": stop_loss,
            "tpTriggerBy": tp_trigger_by,
            "slTriggerBy": sl_trigger_by,
            "reduceOnly": reduce_only,
            "closeOnTrigger": close_on_trigger,
            "smpType": smp_type,
            "mmp": mmp,
            "tpslMode": tpsl_mode,
            "tpLimitPrice": tp_limit_price,
            "slLimitPrice": sl_limit_price,
            "tpOrderType": tp_order_type,
            "slOrderType": sl_order_type,
        }

        return await self._make_request("POST", "/v5/order/create", params=params, signed=True)

    async def amend_order(
        self,
        category: Literal["linear", "inverse", "spot", "option"],
        symbol: str,
        order_id: str | None = None,
        order_link_id: str | None = None,
        order_iv: str | None = None,
        trigger_price: str | None = None,
        qty: str | None = None,
        price: str | None = None,
        tpsl_mode: Literal["Full", "Partial"] | None = None,
        take_profit: str | None = None,
        stop_loss: str | None = None,
        tp_trigger_by: str | None = None,
        sl_trigger_by: str | None = None,
        trigger_by: str | None = None,
        tp_limit_price: str | None = None,
        sl_limit_price: str | None = None,
    ) -> dict:
        """Изменение параметров ордера.

        https://bybit-exchange.github.io/docs/v5/order/amend-order
        """
        params = {
            "category": category,
            "symbol": symbol,
            "orderId": order_id,
            "orderLinkId": order_link_id,
            "orderIv": order_iv,
            "triggerPrice": trigger_price,
            "qty": qty,
            "price": price,
            "tpslMode": tpsl_mode,
            "takeProfit": take_profit,
            "stopLoss": stop_loss,
            "tpTriggerBy": tp_trigger_by,
            "slTriggerBy": sl_trigger_by,
            "triggerBy": trigger_by,
            "tpLimitPrice": tp_limit_price,
            "slLimitPrice": sl_limit_price,
        }

        return await self._make_request("POST", "/v5/order/amend", params=params, signed=True)

    async def cancel_order(
        self,
        category: Literal["linear", "inverse", "spot", "option"],
        symbol: str,
        order_id: str | None = None,
        order_link_id: str | None = None,
        order_filter: str | None = None,
    ) -> dict:
        """Отмена ордера.

        https://bybit-exchange.github.io/docs/v5/order/cancel-order
        """
        params = {
            "category": category,
            "symbol": symbol,
            "orderId": order_id,
            "orderLinkId": order_link_id,
            "orderFilter": order_filter,
        }

        return await self._make_request("POST", "/v5/order/cancel", params=params, signed=True)

    async def open_orders(
        self,
        category: Literal["linear", "inverse", "spot", "option"],
        symbol: str | None = None,
        base_coin: str | None = None,
        settle_coin: str | None = None,
        order_id: str | None = None,
        order_link_id: str | None = None,
        open_only: int | None = None,
        order_filter: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Открытые и закрытые ордера (реaltime).

        https://bybit-exchange.github.io/docs/v5/order/open-order
        """
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
            "orderId": order_id,
            "orderLinkId": order_link_id,
            "openOnly": open_only,
            "orderFilter": order_filter,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request("GET", "/v5/order/realtime", params=params, signed=True)

    async def cancel_all_orders(
        self,
        category: Literal["linear", "inverse", "spot", "option"],
        symbol: str | None = None,
        base_coin: str | None = None,
        settle_coin: str | None = None,
        order_filter: str | None = None,
        stop_order_type: str | None = None,
    ) -> dict:
        """Отмена всех открытых ордеров.

        https://bybit-exchange.github.io/docs/v5/order/cancel-all
        """
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
            "orderFilter": order_filter,
            "stopOrderType": stop_order_type,
        }

        return await self._make_request("POST", "/v5/order/cancel-all", params=params, signed=True)

    async def order_history(
        self,
        category: Literal["linear", "inverse", "spot", "option"],
        symbol: str | None = None,
        base_coin: str | None = None,
        settle_coin: str | None = None,
        order_id: str | None = None,
        order_link_id: str | None = None,
        order_filter: str | None = None,
        order_status: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """История ордеров.

        https://bybit-exchange.github.io/docs/v5/order/order-list
        """
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
            "orderId": order_id,
            "orderLinkId": order_link_id,
            "orderFilter": order_filter,
            "orderStatus": order_status,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request("GET", "/v5/order/history", params=params, signed=True)

    async def trade_history(
        self,
        category: Literal["linear", "inverse", "spot", "option"],
        symbol: str | None = None,
        order_id: str | None = None,
        order_link_id: str | None = None,
        base_coin: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        exec_type: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """История сделок (execution list).

        https://bybit-exchange.github.io/docs/v5/order/execution
        """
        params = {
            "category": category,
            "symbol": symbol,
            "orderId": order_id,
            "orderLinkId": order_link_id,
            "baseCoin": base_coin,
            "startTime": start_time,
            "endTime": end_time,
            "execType": exec_type,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request("GET", "/v5/execution/list", params=params, signed=True)

    async def create_orders_batch(
        self,
        category: Literal["linear", "option", "spot", "inverse"],
        request: list[dict[str, Any]],
    ) -> dict:
        """Массовое создание ордеров.

        https://bybit-exchange.github.io/docs/v5/order/batch-place
        """
        params = {
            "category": category,
            "request": request,
        }

        return await self._make_request(
            "POST", "/v5/order/create-batch", params=params, signed=True
        )

    async def amend_orders_batch(
        self,
        category: Literal["linear", "option", "spot", "inverse"],
        request: list[dict[str, Any]],
    ) -> dict:
        """Массовое изменение ордеров.

        https://bybit-exchange.github.io/docs/v5/order/batch-amend
        """
        params = {
            "category": category,
            "request": request,
        }

        return await self._make_request("POST", "/v5/order/amend-batch", params=params, signed=True)

    async def cancel_orders_batch(
        self,
        category: Literal["linear", "option", "spot", "inverse"],
        request: list[dict[str, Any]],
    ) -> dict:
        """Массовая отмена ордеров.

        https://bybit-exchange.github.io/docs/v5/order/batch-cancel
        """
        params = {
            "category": category,
            "request": request,
        }

        return await self._make_request(
            "POST", "/v5/order/cancel-batch", params=params, signed=True
        )

    async def spot_borrow_quota(
        self,
        category: Literal["spot"],
        symbol: str,
        side: str,
    ) -> dict:
        """Доступная квота заимствования (Spot).

        https://bybit-exchange.github.io/docs/v5/order/spot-borrow-quota
        """
        params = {
            "category": category,
            "symbol": symbol,
            "side": side,
        }

        return await self._make_request(
            "GET", "/v5/order/spot-borrow-check", params=params, signed=True
        )

    async def set_disconnect_cancel_all(
        self,
        time_window: int,
        product: Literal["OPTIONS", "DERIVATIVES", "SPOT"] | None = None,
    ) -> dict:
        """Настройка DCP (отмена ордеров при разрыве соединения).

        https://bybit-exchange.github.io/docs/v5/order/dcp
        """
        params = {
            "product": product,
            "timeWindow": time_window,
        }

        return await self._make_request(
            "POST", "/v5/order/disconnected-cancel-all", params=params, signed=True
        )

    async def pre_check_order(
        self,
        category: Literal["linear", "inverse", "option"],
        symbol: str,
        side: str,
        order_type: str,
        qty: str,
        price: str | None = None,
        trigger_direction: int | None = None,
        trigger_price: str | None = None,
        trigger_by: str | None = None,
        order_iv: str | None = None,
        time_in_force: str | None = None,
        position_idx: int | None = None,
        order_link_id: str | None = None,
        take_profit: str | None = None,
        stop_loss: str | None = None,
        tp_trigger_by: str | None = None,
        sl_trigger_by: str | None = None,
        reduce_only: bool | None = None,
        close_on_trigger: bool | None = None,
        smp_type: str | None = None,
        mmp: bool | None = None,
        tpsl_mode: Literal["Full", "Partial"] | None = None,
        tp_limit_price: str | None = None,
        sl_limit_price: str | None = None,
        tp_order_type: str | None = None,
        sl_order_type: str | None = None,
    ) -> dict:
        """Предварительная проверка ордера (IMR/MMR).

        https://bybit-exchange.github.io/docs/v5/order/pre-check-order
        """
        params = {
            "category": category,
            "symbol": symbol,
            "side": side,
            "orderType": order_type,
            "qty": qty,
            "price": price,
            "triggerDirection": trigger_direction,
            "triggerPrice": trigger_price,
            "triggerBy": trigger_by,
            "orderIv": order_iv,
            "timeInForce": time_in_force,
            "positionIdx": position_idx,
            "orderLinkId": order_link_id,
            "takeProfit": take_profit,
            "stopLoss": stop_loss,
            "tpTriggerBy": tp_trigger_by,
            "slTriggerBy": sl_trigger_by,
            "reduceOnly": reduce_only,
            "closeOnTrigger": close_on_trigger,
            "smpType": smp_type,
            "mmp": mmp,
            "tpslMode": tpsl_mode,
            "tpLimitPrice": tp_limit_price,
            "slLimitPrice": sl_limit_price,
            "tpOrderType": tp_order_type,
            "slOrderType": sl_order_type,
        }

        return await self._make_request("POST", "/v5/order/pre-check", params=params, signed=True)

    # topic: position

    async def position_info(
        self,
        category: Literal["linear", "inverse", "option"],
        symbol: str | None = None,
        base_coin: str | None = None,
        settle_coin: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Информация по позициям в реальном времени.

        https://bybit-exchange.github.io/docs/v5/position
        """
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "settleCoin": settle_coin,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request("GET", "/v5/position/list", params=params, signed=True)

    async def set_leverage(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        buy_leverage: str,
        sell_leverage: str,
    ) -> dict:
        """Установка плеча для символа.

        https://bybit-exchange.github.io/docs/v5/position/leverage
        """
        params = {
            "category": category,
            "symbol": symbol,
            "buyLeverage": buy_leverage,
            "sellLeverage": sell_leverage,
        }

        return await self._make_request(
            "POST", "/v5/position/set-leverage", params=params, signed=True
        )

    async def switch_isolated_margin(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        trade_mode: int,
        buy_leverage: str,
        sell_leverage: str,
    ) -> dict:
        """Переключение кросс/изолированной маржи для символа.

        https://bybit-exchange.github.io/docs/v5/position/cross-isolate
        """
        params = {
            "category": category,
            "symbol": symbol,
            "tradeMode": trade_mode,
            "buyLeverage": buy_leverage,
            "sellLeverage": sell_leverage,
        }

        return await self._make_request(
            "POST", "/v5/position/switch-isolated", params=params, signed=True
        )

    async def switch_position_mode(
        self,
        category: Literal["linear", "inverse"],
        mode: int,
        symbol: str | None = None,
        coin: str | None = None,
    ) -> dict:
        """Переключение режима позиций (one-way/hedge).

        https://bybit-exchange.github.io/docs/v5/position/position-mode
        """
        params = {
            "category": category,
            "symbol": symbol,
            "coin": coin,
            "mode": mode,
        }

        return await self._make_request(
            "POST", "/v5/position/switch-mode", params=params, signed=True
        )

    async def set_trading_stop(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        tpsl_mode: Literal["Full", "Partial"],
        position_idx: int,
        take_profit: str | None = None,
        stop_loss: str | None = None,
        trailing_stop: str | None = None,
        tp_trigger_by: str | None = None,
        sl_trigger_by: str | None = None,
        active_price: str | None = None,
        tp_size: str | None = None,
        sl_size: str | None = None,
        tp_limit_price: str | None = None,
        sl_limit_price: str | None = None,
        tp_order_type: str | None = None,
        sl_order_type: str | None = None,
    ) -> dict:
        """Установка TP/SL/Trailing Stop для позиции.

        https://bybit-exchange.github.io/docs/v5/position/trading-stop
        """
        params = {
            "category": category,
            "symbol": symbol,
            "tpslMode": tpsl_mode,
            "positionIdx": position_idx,
            "takeProfit": take_profit,
            "stopLoss": stop_loss,
            "trailingStop": trailing_stop,
            "tpTriggerBy": tp_trigger_by,
            "slTriggerBy": sl_trigger_by,
            "activePrice": active_price,
            "tpSize": tp_size,
            "slSize": sl_size,
            "tpLimitPrice": tp_limit_price,
            "slLimitPrice": sl_limit_price,
            "tpOrderType": tp_order_type,
            "slOrderType": sl_order_type,
        }

        return await self._make_request(
            "POST", "/v5/position/trading-stop", params=params, signed=True
        )

    async def set_auto_add_margin(
        self,
        category: Literal["linear"],
        symbol: str,
        auto_add_margin: int,
        position_idx: int | None = None,
    ) -> dict:
        """Включить/выключить авто-добавление маржи (изолированная позиция).

        https://bybit-exchange.github.io/docs/v5/position/auto-add-margin
        """
        params = {
            "category": category,
            "symbol": symbol,
            "autoAddMargin": auto_add_margin,
            "positionIdx": position_idx,
        }

        return await self._make_request(
            "POST", "/v5/position/set-auto-add-margin", params=params, signed=True
        )

    async def add_margin(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        margin: str,
        position_idx: int | None = None,
    ) -> dict:
        """Ручное добавление/уменьшение маржи (изолированная позиция).

        https://bybit-exchange.github.io/docs/v5/position/manual-add-margin
        """
        params = {
            "category": category,
            "symbol": symbol,
            "margin": margin,
            "positionIdx": position_idx,
        }

        return await self._make_request(
            "POST", "/v5/position/add-margin", params=params, signed=True
        )

    async def closed_pnl(
        self,
        category: Literal["linear", "inverse", "option"],
        symbol: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """История закрытой прибыли/убытка (PnL).

        https://bybit-exchange.github.io/docs/v5/position/close-pnl
        """
        params = {
            "category": category,
            "symbol": symbol,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/position/closed-pnl", params=params, signed=True
        )

    async def closed_option_positions(
        self,
        category: Literal["option"],
        symbol: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Закрытые опционные позиции.

        https://bybit-exchange.github.io/docs/v5/position/close-position
        """
        params = {
            "category": category,
            "symbol": symbol,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/position/get-closed-positions", params=params, signed=True
        )

    async def move_positions(
        self,
        from_uid: str,
        to_uid: str,
        legs: list[dict[str, Any]],
    ) -> dict:
        """Перемещение позиций между учетными записями.

        https://bybit-exchange.github.io/docs/v5/position/move-position
        """
        params = {
            "fromUid": from_uid,
            "toUid": to_uid,
            "list": legs,
        }

        return await self._make_request(
            "POST", "/v5/position/move-positions", params=params, signed=True
        )

    async def move_position_history(
        self,
        category: Literal["linear", "spot", "option"] | None = None,
        symbol: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        status: str | None = None,
        block_trade_id: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """История перемещения позиций.

        https://bybit-exchange.github.io/docs/v5/position/move-position-history
        """
        params = {
            "category": category,
            "symbol": symbol,
            "startTime": start_time,
            "endTime": end_time,
            "status": status,
            "blockTradeId": block_trade_id,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/position/move-history", params=params, signed=True
        )

    async def confirm_pending_mmr(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
    ) -> dict:
        """Подтверждение нового уровня риска (MMR) для снятия reduceOnly.

        https://bybit-exchange.github.io/docs/v5/position/confirm-mmr
        """
        params = {
            "category": category,
            "symbol": symbol,
        }

        return await self._make_request(
            "POST", "/v5/position/confirm-pending-mmr", params=params, signed=True
        )

    async def set_tpsl_mode(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        tp_sl_mode: str,
    ) -> dict:
        """Установка режима TP/SL по умолчанию для символа.

        https://bybit-exchange.github.io/docs/v5/position/tpsl-mode
        """
        params = {
            "category": category,
            "symbol": symbol,
            "tpSlMode": tp_sl_mode,
        }

        return await self._make_request(
            "POST", "/v5/position/set-tpsl-mode", params=params, signed=True
        )

    async def set_risk_limit(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        risk_id: int,
        position_idx: int | None = None,
    ) -> dict:
        """Установка лимита риска позиции.

        https://bybit-exchange.github.io/docs/v5/position/set-risk-limit
        """
        params = {
            "category": category,
            "symbol": symbol,
            "riskId": risk_id,
            "positionIdx": position_idx,
        }

        return await self._make_request(
            "POST", "/v5/position/set-risk-limit", params=params, signed=True
        )

    # topic: pre-upgrade

    async def pre_upgrade_order_history(
        self,
        category: Literal["linear", "inverse", "option", "spot"],
        symbol: str | None = None,
        base_coin: str | None = None,
        order_id: str | None = None,
        order_link_id: str | None = None,
        order_filter: str | None = None,
        order_status: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """История ордеров до апгрейда аккаунта.

        https://bybit-exchange.github.io/docs/v5/pre-upgrade/order-list
        """
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
            "orderId": order_id,
            "orderLinkId": order_link_id,
            "orderFilter": order_filter,
            "orderStatus": order_status,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/pre-upgrade/order/history", params=params, signed=True
        )

    async def pre_upgrade_trade_history(
        self,
        category: Literal["linear", "inverse", "option", "spot"],
        symbol: str | None = None,
        order_id: str | None = None,
        order_link_id: str | None = None,
        base_coin: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        exec_type: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """История сделок до апгрейда аккаунта.

        https://bybit-exchange.github.io/docs/v5/pre-upgrade/execution
        """
        params = {
            "category": category,
            "symbol": symbol,
            "orderId": order_id,
            "orderLinkId": order_link_id,
            "baseCoin": base_coin,
            "startTime": start_time,
            "endTime": end_time,
            "execType": exec_type,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/pre-upgrade/execution/list", params=params, signed=True
        )

    async def pre_upgrade_closed_pnl(
        self,
        category: Literal["linear", "inverse"],
        symbol: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Закрытый PnL до апгрейда аккаунта.

        https://bybit-exchange.github.io/docs/v5/pre-upgrade/close-pnl
        """
        params = {
            "category": category,
            "symbol": symbol,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/pre-upgrade/position/closed-pnl", params=params, signed=True
        )

    async def pre_upgrade_transaction_log(
        self,
        category: Literal["linear", "option"],
        base_coin: str | None = None,
        type: str | None = None,  # noqa: A003 - API param name
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Транзакционный лог USDC деривативов до апгрейда аккаунта.

        https://bybit-exchange.github.io/docs/v5/pre-upgrade/transaction-log
        """
        params = {
            "category": category,
            "baseCoin": base_coin,
            "type": type,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/pre-upgrade/account/transaction-log", params=params, signed=True
        )

    async def pre_upgrade_delivery_record(
        self,
        category: Literal["option"],
        symbol: str | None = None,
        exp_date: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Опционные записи поставок до апгрейда аккаунта.

        https://bybit-exchange.github.io/docs/v5/pre-upgrade/delivery
        """
        params = {
            "category": category,
            "symbol": symbol,
            "expDate": exp_date,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/pre-upgrade/asset/delivery-record", params=params, signed=True
        )

    async def pre_upgrade_settlement_record(
        self,
        category: Literal["linear"],
        symbol: str | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Сессионные расчёты USDC Perpetual до апгрейда аккаунта.

        https://bybit-exchange.github.io/docs/v5/pre-upgrade/settlement
        """
        params = {
            "category": category,
            "symbol": symbol,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/pre-upgrade/asset/settlement-record", params=params, signed=True
        )

    # topic: account

    async def wallet_balance(
        self,
        account_type: Literal["UNIFIED", "CONTRACT", "SPOT"],
        coin: str | None = None,
    ) -> dict:
        """Баланс кошелька и активы по монетам.

        https://bybit-exchange.github.io/docs/v5/account/wallet-balance
        """
        params = {
            "accountType": account_type,
            "coin": coin,
        }

        return await self._make_request(
            "GET", "/v5/account/wallet-balance", params=params, signed=True
        )

    async def transferable_amount(
        self,
        coin_name: str,
    ) -> dict:
        """Доступная к переводу сумма (Unified).

        https://bybit-exchange.github.io/docs/v5/account/unified-trans-amnt
        """
        params = {
            "coinName": coin_name,
        }

        return await self._make_request("GET", "/v5/account/withdrawal", params=params, signed=True)

    async def upgrade_to_unified_account(self) -> dict:
        """Апгрейд до Unified аккаунта.

        https://bybit-exchange.github.io/docs/v5/account/upgrade-unified-account
        """
        return await self._make_request(
            "POST", "/v5/account/upgrade-to-uta", params={}, signed=True
        )

    async def borrow_history(
        self,
        currency: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """История заимствований (проценты).

        https://bybit-exchange.github.io/docs/v5/account/borrow-history
        """
        params = {
            "currency": currency,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/account/borrow-history", params=params, signed=True
        )

    async def quick_repayment(
        self,
        coin: str | None = None,
    ) -> dict:
        """Погашение обязательств Unified аккаунта.

        https://bybit-exchange.github.io/docs/v5/account/repay-liability
        """
        params = {
            "coin": coin,
        }

        return await self._make_request(
            "POST", "/v5/account/quick-repayment", params=params, signed=True
        )

    async def set_collateral_coin(
        self,
        coin: str,
        collateral_switch: Literal["ON", "OFF"],
    ) -> dict:
        """Настройка монеты как залога.

        https://bybit-exchange.github.io/docs/v5/account/set-collateral
        """
        params = {
            "coin": coin,
            "collateralSwitch": collateral_switch,
        }

        return await self._make_request(
            "POST", "/v5/account/set-collateral-switch", params=params, signed=True
        )

    async def set_collateral_coin_batch(
        self,
        request: list[dict[str, Any]],
    ) -> dict:
        """Пакетная настройка монет как залога.

        https://bybit-exchange.github.io/docs/v5/account/batch-set-collateral
        """
        params = {
            "request": request,
        }

        return await self._make_request(
            "POST", "/v5/account/set-collateral-switch-batch", params=params, signed=True
        )

    async def collateral_info(
        self,
        currency: str | None = None,
    ) -> dict:
        """Информация о залоге (процентные ставки, коэффициенты и пр.).

        https://bybit-exchange.github.io/docs/v5/account/collateral-info
        """
        params = {
            "currency": currency,
        }

        return await self._make_request(
            "GET", "/v5/account/collateral-info", params=params, signed=True
        )

    async def coin_greeks(
        self,
        base_coin: str | None = None,
    ) -> dict:
        """Текущие греческие параметры аккаунта по базовым монетам.

        https://bybit-exchange.github.io/docs/v5/account/coin-greeks
        """
        params = {
            "baseCoin": base_coin,
        }

        return await self._make_request("GET", "/v5/asset/coin-greeks", params=params, signed=True)

    async def fee_rate(
        self,
        category: Literal["spot", "linear", "inverse", "option"],
        symbol: str | None = None,
        base_coin: str | None = None,
    ) -> dict:
        """Торговые комиссии аккаунта.

        https://bybit-exchange.github.io/docs/v5/account/fee-rate
        """
        params = {
            "category": category,
            "symbol": symbol,
            "baseCoin": base_coin,
        }

        return await self._make_request("GET", "/v5/account/fee-rate", params=params, signed=True)

    async def account_info(self) -> dict:
        """Информация об аккаунте (режимы маржи и т.п.).

        https://bybit-exchange.github.io/docs/v5/account/account-info
        """
        return await self._make_request("GET", "/v5/account/info", params={}, signed=True)

    async def dcp_info(self) -> dict:
        """Конфигурация DCP аккаунта.

        https://bybit-exchange.github.io/docs/v5/account/dcp-info
        """
        return await self._make_request("GET", "/v5/account/query-dcp-info", params={}, signed=True)

    async def transaction_log(
        self,
        account_type: Literal["UNIFIED"] | None = None,
        category: Literal["spot", "linear", "option", "inverse"] | None = None,
        currency: str | None = None,
        base_coin: str | None = None,
        type: str | None = None,  # noqa: A003 - API param name
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Транзакционный лог Unified аккаунта.

        https://bybit-exchange.github.io/docs/v5/account/transaction-log
        """
        params = {
            "accountType": account_type,
            "category": category,
            "currency": currency,
            "baseCoin": base_coin,
            "type": type,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/account/transaction-log", params=params, signed=True
        )

    async def contract_transaction_log(
        self,
        currency: str | None = None,
        base_coin: str | None = None,
        type: str | None = None,  # noqa: A003 - API param name
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        cursor: str | None = None,
    ) -> dict:
        """Транзакционный лог деривативного кошелька.

        https://bybit-exchange.github.io/docs/v5/account/contract-transaction-log
        """
        params = {
            "currency": currency,
            "baseCoin": base_coin,
            "type": type,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "cursor": cursor,
        }

        return await self._make_request(
            "GET", "/v5/account/contract-transaction-log", params=params, signed=True
        )

    async def smp_group(self) -> dict:
        """Получить SMP group ID.

        https://bybit-exchange.github.io/docs/v5/account/smp-group
        """
        return await self._make_request("GET", "/v5/account/smp-group", params={}, signed=True)

    async def set_margin_mode(
        self,
        set_margin_mode: Literal["ISOLATED_MARGIN", "REGULAR_MARGIN", "PORTFOLIO_MARGIN"],
    ) -> dict:
        """Установить режим маржи аккаунта.

        https://bybit-exchange.github.io/docs/v5/account/set-margin-mode
        """
        params = {
            "setMarginMode": set_margin_mode,
        }

        return await self._make_request(
            "POST", "/v5/account/set-margin-mode", params=params, signed=True
        )

    async def set_spot_hedging(
        self,
        set_hedging_mode: Literal["ON", "OFF"],
    ) -> dict:
        """Включить/выключить спотовый hedging в Portfolio margin.

        https://bybit-exchange.github.io/docs/v5/account/set-spot-hedge
        """
        params = {
            "setHedgingMode": set_hedging_mode,
        }

        return await self._make_request(
            "POST", "/v5/account/set-hedging-mode", params=params, signed=True
        )

    async def set_limit_price_action(
        self,
        category: Literal["linear", "inverse", "spot"],
        modify_enable: bool,
    ) -> dict:
        """Настройка поведения при выходе лимитной цены за границы.

        https://bybit-exchange.github.io/docs/v5/account/set-price-limit
        """
        params = {
            "category": category,
            "modifyEnable": modify_enable,
        }

        return await self._make_request(
            "POST", "/v5/account/set-limit-px-action", params=params, signed=True
        )

    async def user_setting_config(self) -> dict:
        """Получить конфигурацию поведения для лимитных цен.

        https://bybit-exchange.github.io/docs/v5/account/get-user-setting-config
        """
        return await self._make_request(
            "GET", "/v5/account/user-setting-config", params={}, signed=True
        )

    async def set_mmp(
        self,
        base_coin: str,
        window: str,
        frozen_period: str,
        qty_limit: str,
        delta_limit: str,
    ) -> dict:
        """Настроить Market Maker Protection (MMP).

        https://bybit-exchange.github.io/docs/v5/account/set-mmp
        """
        params = {
            "baseCoin": base_coin,
            "window": window,
            "frozenPeriod": frozen_period,
            "qtyLimit": qty_limit,
            "deltaLimit": delta_limit,
        }

        return await self._make_request(
            "POST", "/v5/account/mmp-modify", params=params, signed=True
        )

    async def reset_mmp(
        self,
        base_coin: str,
    ) -> dict:
        """Сбросить состояние MMP (разморозить).

        https://bybit-exchange.github.io/docs/v5/account/reset-mmp
        """
        params = {
            "baseCoin": base_coin,
        }

        return await self._make_request("POST", "/v5/account/mmp-reset", params=params, signed=True)

    async def mmp_state(
        self,
        base_coin: str,
    ) -> dict:
        """Статус MMP.

        https://bybit-exchange.github.io/docs/v5/account/get-mmp-state
        """
        params = {
            "baseCoin": base_coin,
        }

        return await self._make_request("GET", "/v5/account/mmp-state", params=params, signed=True)



__all__ = ["Client"]


from typing import Any

from unicex._base import BaseClient
from unicex.types import RequestMethod
from unicex.utils import filter_params, get_timestamp


class Client(BaseClient):
    """Клиент для работы с BingX API."""

    _BASE_URL: str = "https://open-api.bingx.com"
    """Базовый URL для REST API BingX."""

    async def _make_request(
        self,
        method: RequestMethod,
        endpoint: str,
        *,
        params: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Выполняет HTTP-запрос к эндпоинтам BingX API.

        Параметры:
            method (str): HTTP метод запроса ("GET", "POST", "DELETE" и т.д.).
            endpoint (str): URL эндпоинта Kucoin API.
            params (dict | None): Параметры запроса.

        Возвращает:
            dict: Ответ в формате JSON.
        """
        # Составляем URL для запроса
        url = self._BASE_URL + endpoint

        # Добавляем timestap в параметры, если он не указан
        if params and "timestamp" in params and not params["timestamp"]:
            params["timestamp"] = get_timestamp()

        # Фильтруем параметры от None значений
        params = filter_params(params) if params else {}

        # Выполняем запрос
        return await super()._make_request(
            method=method,
            url=url,
            params=params,
        )

    async def futures_contracts(
        self,
        symbol: str | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение списка USDT-M бессрочных фьючерсных контрактов.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/USDT-M%20Perp%20Futures%20symbols
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v2/quote/contracts", params=params)

    async def futures_order_book(
        self,
        symbol: str,
        limit: int | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение стакана ордеров.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Order%20Book
        """
        params = {
            "symbol": symbol,
            "limit": limit,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v2/quote/depth", params=params)

    async def futures_trades(
        self,
        symbol: str,
        limit: int | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение списка последних сделок.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Recent%20Trades%20List
        """
        params = {
            "symbol": symbol,
            "limit": limit,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v2/quote/trades", params=params)

    async def futures_mark_price(
        self,
        symbol: str | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение цены маркировки и ставки финансирования.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Mark%20Price%20and%20Funding%20Rate
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v2/quote/premiumIndex", params=params)

    async def futures_funding_rate_history(
        self,
        symbol: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение истории ставок финансирования.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Get%20Funding%20Rate%20History
        """
        params = {
            "symbol": symbol,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v2/quote/fundingRate", params=params)

    async def futures_klines(
        self,
        symbol: str,
        interval: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение свечей.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Kline%2FCandlestick%20Data
        """
        params = {
            "symbol": symbol,
            "interval": interval,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v3/quote/klines", params=params)

    async def open_interest(
        self,
        symbol: str,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение статистики открытого интереса.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Open%20Interest%20Statistics
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v2/quote/openInterest", params=params)

    async def futures_ticker_24hr(
        self,
        symbol: str | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение 24-часовой статистики тикера.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/24hr%20Ticker%20Price%20Change%20Statistics
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v2/quote/ticker", params=params)

    async def futures_historical_trades(
        self,
        symbol: str | None = None,
        from_id: int | None = None,
        limit: int | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение исторических сделок.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Query%20historical%20transaction%20orders
        """
        params = {
            "fromId": from_id,
            "symbol": symbol,
            "limit": limit,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request(
            "GET",
            "/openApi/swap/v1/market/historicalTrades",
            params=params,
        )

    async def futures_book_ticker(
        self,
        symbol: str,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение лучшего бид/аск по символу.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Symbol%20Order%20Book%20Ticker
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v2/quote/bookTicker", params=params)

    async def futures_mark_price_klines(
        self,
        symbol: str,
        interval: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение свечей цены маркировки.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Mark%20Price%20Kline-Candlestick%20Data
        """
        params = {
            "symbol": symbol,
            "interval": interval,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request(
            "GET",
            "/openApi/swap/v1/market/markPriceKlines",
            params=params,
        )

    async def futures_ticker_price(
        self,
        symbol: str | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение последней цены тикера.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Symbol%20Price%20Ticker
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v1/ticker/price", params=params)

    async def futures_trading_rules(
        self,
        symbol: str | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение торговых правил.

        https://bingx-api.github.io/docs-v3/#/en/Swap/Market%20Data/Trading%20Rules
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/swap/v1/tradingRules", params=params)

    async def symbols(
        self,
        symbol: str | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение списка спотовых торговых пар.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/Spot%20trading%20symbols
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/spot/v1/common/symbols", params=params)

    async def trades(
        self,
        symbol: str,
        limit: int | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение списка последних сделок.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/Recent%20Trades%20List
        """
        params = {
            "symbol": symbol,
            "limit": limit,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/spot/v1/market/trades", params=params)

    async def order_book(
        self,
        symbol: str,
        limit: int | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение стакана ордеров.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/Order%20Book
        """
        params = {
            "symbol": symbol,
            "limit": limit,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/spot/v1/market/depth", params=params)

    async def klines(
        self,
        symbol: str,
        interval: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение свечей.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/Kline-Candlestick%20Data
        """
        params = {
            "symbol": symbol,
            "interval": interval,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/spot/v2/market/kline", params=params)

    async def ticker_24hr(
        self,
        symbol: str | None = None,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение 24-часовой статистики тикера.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/24hr%20Ticker%20Price%20Change%20Statistics
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/spot/v1/ticker/24hr", params=params)

    async def order_book_agg(
        self,
        symbol: str,
        depth: int,
        type_: str,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение агрегированного стакана ордеров.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/Order%20Book%20aggregation
        """
        params = {
            "symbol": symbol,
            "depth": depth,
            "type": type_,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/spot/v2/market/depth", params=params)

    async def ticker_price(
        self,
        symbol: str,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение последней цены тикера.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/Symbol%20Price%20Ticker
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/spot/v2/ticker/price", params=params)

    async def book_ticker(
        self,
        symbol: str,
        timestamp: int | None = None,
        recv_window: int | None = None,
    ) -> dict[str, Any]:
        """Получение лучшего бид/аск по символу.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/Symbol%20Order%20Book%20Ticker
        """
        params = {
            "symbol": symbol,
            "timestamp": timestamp,
            "recvWindow": recv_window,
        }

        return await self._make_request("GET", "/openApi/spot/v1/ticker/bookTicker", params=params)

    async def historical_klines(
        self,
        symbol: str,
        interval: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
    ) -> dict[str, Any]:
        """Получение исторических свечей.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/Historical%20K-line
        """
        params = {
            "symbol": symbol,
            "interval": interval,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
        }

        return await self._make_request("GET", "/openApi/market/his/v1/kline", params=params)

    async def historical_trades(
        self,
        symbol: str,
        limit: int | None = None,
        from_id: str | None = None,
    ) -> dict[str, Any]:
        """Получение исторических сделок.

        https://bingx-api.github.io/docs-v3/#/en/Spot/Market%20Data/Old%20Trade%20Lookup
        """
        params = {
            "symbol": symbol,
            "limit": limit,
            "fromId": from_id,
        }

        return await self._make_request("GET", "/openApi/market/his/v1/trade", params=params)



__all__ = ["WebsocketManager"]


import json
from collections.abc import Awaitable, Callable
from typing import Any, Literal

from unicex._base import Websocket

from .client import Client

type CallbackType = Callable[[Any], Awaitable[None]]


class WebsocketManager:
    """Менеджер асинхронных вебсокетов для Okx."""

    _BASE_URL: str = "wss://ws.okx.com:8443/ws"
    """Базовый URL вебсокетов на Okx."""

    _PUBLIC_URL: str = _BASE_URL + "/v5/public"
    """Публичный URL вебсокетов на Okx."""

    _BUSINESS_URL: str = _BASE_URL + "/v5/business"
    """Бизнес-URL вебсокетов на Okx. (для топиков trades-all и candle)"""

    def __init__(self, client: Client | None = None, **ws_kwargs: Any) -> None:
        """Инициализирует менеджер вебсокетов для OKX.

        Параметры:
            client (`Client | None`): Клиент для выполнения запросов. Нужен, чтобы открыть приватные вебсокеты.
            ws_kwargs (`dict[str, Any]`): Дополнительные аргументы, которые прокидываются в `Websocket`.
        """
        self.client = client
        self._ws_kwargs = {"ping_message": "ping", **ws_kwargs}

    def _build_subscription_message(self, args: list[dict[str, Any]]) -> str:
        """Формирует JSON-сообщение подписки."""
        return json.dumps(
            {
                "op": "subscribe",
                "args": args,
            }
        )

    def _normalize_inst_ids(self, inst_id: str | list[str]) -> list[str]:
        """Нормализует inst_id до списка."""
        if isinstance(inst_id, str):
            return [inst_id]
        if not inst_id:
            raise ValueError("inst_id list cannot be empty")
        return list(inst_id)

    def _build_inst_id_args(
        self,
        base_args: dict[str, str],
        inst_id: str | list[str],
    ) -> list[dict[str, str]]:
        """Формирует args для списка inst_id."""
        inst_ids = self._normalize_inst_ids(inst_id)
        return [{**base_args, "instId": item} for item in inst_ids]

    def instruments(
        self,
        callback: CallbackType,
        inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"],
    ) -> Websocket:
        """Создает вебсокет для получения изменений в состоянии инструментов.

        https://www.okx.com/docs-v5/en/#public-data-websocket-instruments-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_type (`Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"]`): Тип инструмента.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._build_subscription_message(
            [
                {
                    "channel": "instruments",
                    "instType": inst_type,
                }
            ]
        )

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def open_interest(self, callback: CallbackType, inst_id: str | list[str]) -> Websocket:
        """Создает вебсокет для получения данных об открытом интересе.

        https://www.okx.com/docs-v5/en/#public-data-websocket-open-interest-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "LTC-USD-SWAP").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {"channel": "open-interest"},
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def funding_rate(
        self,
        callback: CallbackType,
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения данных о ставке финансирования.

        https://www.okx.com/docs-v5/en/#public-data-websocket-funding-rate-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "BTC-USD-SWAP").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {"channel": "funding-rate"},
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def price_limit(
        self,
        callback: CallbackType,
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения максимальной цены покупки и минимальной цены продажи инструментов.

        https://www.okx.com/docs-v5/en/#public-data-websocket-price-limit-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "LTC-USD-190628").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {"channel": "price-limit"},
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def option_summary(
        self,
        callback: CallbackType,
        inst_family: str,
    ) -> Websocket:
        """Создает вебсокет для получения детальной информации о ценах всех OPTION контрактов.

        https://www.okx.com/docs-v5/en/#public-data-websocket-option-summary-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_family (`str`): Семейство инструментов (например, "BTC-USD").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._build_subscription_message(
            [
                {
                    "channel": "opt-summary",
                    "instFamily": inst_family,
                }
            ]
        )

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def estimated_price(
        self,
        callback: CallbackType,
        inst_type: Literal["OPTION", "FUTURES"],
        inst_family: str | None = None,
        inst_id: str | list[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения расчетной цены поставки/исполнения/расчета для FUTURES и OPTION контрактов.

        https://www.okx.com/docs-v5/en/#public-data-websocket-estimated-delivery-exercise-settlement-price-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_type (`Literal["OPTION", "FUTURES"]`): Тип инструмента.
            inst_family (`str | None`): Семейство инструментов (например, "BTC-USD"). Обязателен либо inst_family, либо inst_id.
            inst_id (`str | list[str] | None`): ID инструмента или список ID. Обязателен либо inst_family, либо inst_id.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        if not inst_family and not inst_id:
            raise ValueError("Either inst_family or inst_id must be provided")
        if inst_family and inst_id:
            raise ValueError("Only one of inst_family or inst_id should be provided")

        base_args: dict[str, str] = {
            "channel": "estimated-price",
            "instType": inst_type,
        }

        if inst_family:
            base_args["instFamily"] = inst_family
            args = [base_args]
        else:
            args = self._build_inst_id_args(base_args, inst_id)  # type: ignore

        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def mark_price(
        self,
        callback: CallbackType,
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения маркировочной цены.

        https://www.okx.com/docs-v5/en/#public-data-websocket-mark-price-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "LTC-USD-190628").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {
                "channel": "mark-price",
            },
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def index_tickers(
        self,
        callback: CallbackType,
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения данных индексных тикеров.

        https://www.okx.com/docs-v5/en/#public-data-websocket-index-tickers-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_id (`str | list[str]`): Индекс или список индексов (например, "BTC-USDT").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {
                "channel": "index-tickers",
            },
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def mark_price_candlesticks(
        self,
        callback: CallbackType,
        interval: Literal[
            "3M",
            "1M",
            "1W",
            "1D",
            "2D",
            "3D",
            "5D",
            "12H",
            "6H",
            "4H",
            "2H",
            "1H",
            "30m",
            "15m",
            "5m",
            "3m",
            "1m",
            "1Yutc",
            "3Mutc",
            "1Mutc",
            "1Wutc",
            "1Dutc",
            "2Dutc",
            "3Dutc",
            "5Dutc",
            "12Hutc",
            "6Hutc",
        ],
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения данных свечей маркировочной цены.

        https://www.okx.com/docs-v5/en/#public-data-websocket-mark-price-candlesticks-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            interval (`Literal`): Интервал свечей.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "BTC-USD-190628").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        channel = f"mark-price-candle{interval}"
        args = self._build_inst_id_args(
            {
                "channel": channel,
            },
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._BUSINESS_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def index_candlesticks(
        self,
        callback: CallbackType,
        interval: Literal[
            "3M",
            "1M",
            "1W",
            "1D",
            "2D",
            "3D",
            "5D",
            "12H",
            "6H",
            "4H",
            "2H",
            "1H",
            "30m",
            "15m",
            "5m",
            "3m",
            "1m",
            "3Mutc",
            "1Mutc",
            "1Wutc",
            "1Dutc",
            "2Dutc",
            "3Dutc",
            "5Dutc",
            "12Hutc",
            "6Hutc",
        ],
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения данных свечей индекса.

        https://www.okx.com/docs-v5/en/#public-data-websocket-index-candlesticks-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            interval (`Literal`): Интервал свечей.
            inst_id (`str | list[str]`): Индекс или список индексов (например, "BTC-USD").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        channel = f"index-candle{interval}"
        args = self._build_inst_id_args(
            {
                "channel": channel,
            },
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._BUSINESS_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def liquidation_orders(
        self,
        callback: CallbackType,
        inst_type: Literal["SWAP", "FUTURES", "MARGIN", "OPTION"],
    ) -> Websocket:
        """Создает вебсокет для получения недавних ордеров ликвидации.

        https://www.okx.com/docs-v5/en/#public-data-websocket-liquidation-orders-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_type (`Literal["SWAP", "FUTURES", "MARGIN", "OPTION"]`): Тип инструмента.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._build_subscription_message(
            [
                {
                    "channel": "liquidation-orders",
                    "instType": inst_type,
                }
            ]
        )

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def adl_warning(
        self,
        callback: CallbackType,
        inst_type: Literal["SWAP", "FUTURES", "OPTION"],
        inst_family: str | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения предупреждений об авто-делевередже.

        https://www.okx.com/docs-v5/en/#public-data-websocket-adl-warning-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_type (`Literal["SWAP", "FUTURES", "OPTION"]`): Тип инструмента.
            inst_family (`str | None`): Семейство инструментов (например, "BTC-USDT").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args: dict[str, str] = {
            "channel": "adl-warning",
            "instType": inst_type,
        }

        if inst_family:
            args["instFamily"] = inst_family

        subscription_message = self._build_subscription_message([args])

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def tickers(
        self,
        callback: CallbackType,
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения последней цены сделки, цены bid, цены ask и 24-часового объема торгов.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-ws-tickers-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "BTC-USDT").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {
                "channel": "tickers",
            },
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def candlesticks(
        self,
        callback: CallbackType,
        interval: Literal[
            "3M",
            "1M",
            "1W",
            "1D",
            "2D",
            "3D",
            "5D",
            "12H",
            "6H",
            "4H",
            "2H",
            "1H",
            "30m",
            "15m",
            "5m",
            "3m",
            "1m",
            "1s",
            "3Mutc",
            "1Mutc",
            "1Wutc",
            "1Dutc",
            "2Dutc",
            "3Dutc",
            "5Dutc",
            "12Hutc",
            "6Hutc",
        ],
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения данных свечей инструмента.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-ws-candlesticks-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            interval (`Literal`): Интервал свечей.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "BTC-USDT").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {
                "channel": f"candle{interval}",
            },
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._BUSINESS_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def trades(
        self,
        callback: CallbackType,
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения данных о последних сделках.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-ws-trades-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "BTC-USDT").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {
                "channel": "trades",
            },
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def all_trades(
        self,
        callback: CallbackType,
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения данных о всех сделках (по одной сделке на обновление).

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-ws-all-trades-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "BTC-USDT").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {
                "channel": "trades-all",
            },
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._BUSINESS_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def order_book(
        self,
        callback: CallbackType,
        channel: Literal["books", "books5", "bbo-tbt", "books50-l2-tbt", "books-l2-tbt"],
        inst_id: str | list[str],
    ) -> Websocket:
        """Создает вебсокет для получения данных ордербука.

        https://www.okx.com/docs-v5/en/#order-book-trading-market-data-ws-order-book-channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            channel (`Literal["books", "books5", "bbo-tbt", "books50-l2-tbt", "books-l2-tbt"]`): Тип канала ордербука.
            inst_id (`str | list[str]`): ID инструмента или список ID (например, "BTC-USDT").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        args = self._build_inst_id_args(
            {
                "channel": channel,
            },
            inst_id,
        )
        subscription_message = self._build_subscription_message(args)

        return Websocket(
            callback=callback,
            url=self._PUBLIC_URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )



__all__ = ["Adapter"]

from typing import Any

from unicex.types import (
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    TickerDailyDict,
    TickerDailyItem,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods

from .exchange_info import ExchangeInfo


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Kucoin API."""

    @staticmethod
    def tickers(raw_data: dict, only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о тикерах в список тикеров.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["symbol"]
            for item in raw_data["data"]["list"]
            if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def futures_tickers(raw_data: dict, only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о тикерах в список тикеров.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["symbol"]
            for item in raw_data["data"]["list"]
            if item["symbol"].endswith("USDTM") or not only_usdt
        ]

    @staticmethod
    def ticker_24hr(raw_data: dict) -> TickerDailyDict:
        """Преобразует сырой ответ, в котором содержатся данные о тикере за последние 24 часа в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            TickerDailyDict: Словарь, где ключ - тикер, а значение - статистика за последние 24 часа.
        """

        def safe_float(value: object, default: float = 0.0) -> float:
            try:
                if value is None:
                    return default
                return float(value)  # type: ignore
            except (TypeError, ValueError):
                return default

        result: dict[str, TickerDailyItem] = {}

        for item in raw_data["data"]["list"]:
            symbol = item.get("symbol")
            if not symbol:
                continue

            last_price = safe_float(item.get("lastPrice"))
            open_price = safe_float(item.get("open"))
            base_volume = safe_float(item.get("baseVolume"))
            quote_volume = safe_float(item.get("quoteVolume"))

            if open_price > 0:
                p = round((last_price / open_price - 1) * 100, 2)
            else:
                p = 0.0

            result[symbol] = TickerDailyItem(
                p=p,
                v=base_volume,
                q=quote_volume,
            )

        return result

    @staticmethod
    def last_price(raw_data: dict) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о последней цене, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        return {
            item["symbol"]: float(item["lastPrice"])
            for item in raw_data["data"]["list"]
            if item["lastPrice"]
        }

    @staticmethod
    def open_interest(raw_data: dict[str, Any]) -> OpenInterestDict:
        """Преобразует сырой ответ, в котором содержатся данные об открытом интересе, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            OpenInterestDict: Словарь, где ключ - тикер, а значение - агрегированные данные открытого интереса.
        """
        return {
            item["symbol"]: OpenInterestItem(
                t=item["ts"],
                v=float(item["openInterest"]) * Adapter._get_contract_size(item["symbol"]),
                u="coins",
            )
            for item in raw_data["data"]
        }

    @staticmethod
    def funding_rate(raw_data: dict) -> float:
        """Преобразует историю ставок финансирования в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            float: Актуальная ставка финансирования.
        """
        return round(raw_data["data"]["nextFundingRate"] * 100, 6)

    @staticmethod
    def _get_contract_size(symbol: str) -> float:
        """Возвращает размер контракта для указанного символа тикера."""
        try:
            return ExchangeInfo.get_futures_ticker_info(symbol)["contract_size"] or 1
        except:  # noqa
            return 1

    @staticmethod
    def klines(raw_data: dict, symbol: str) -> list[KlineDict]:
        """Преобразует данные о свечах в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.
            symbol (str): Символ тикера.

        Возвращает:
            list[KlineDict]: Список свечей.
        """
        klines: list[KlineDict] = []
        for item in sorted(raw_data["data"]["list"], key=lambda x: int(float(x[0]))):
            klines.append(  # noqa: PERF401
                KlineDict(
                    s=symbol,
                    t=item[0],
                    o=float(item[1]),
                    h=float(item[3]),
                    l=float(item[4]),
                    c=float(item[2]),
                    v=float(item[5]),
                    q=float(item[6]),
                    T=None,
                    x=None,
                )
            )
        return klines



__all__ = ["Adapter"]

from typing import Any

from unicex.types import (
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods

from .exchange_info import ExchangeInfo


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Mexc API."""

    @staticmethod
    def tickers(raw_data: list[dict], only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о тикерах, в список тикеров.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["symbol"] for item in raw_data if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def futures_tickers(raw_data: dict, only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о фьючерсных тикерах, в список тикеров.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["symbol"]
            for item in raw_data["data"]
            if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def last_price(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о спотовых ценах, в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        return {item["symbol"]: float(item["lastPrice"]) for item in raw_data}

    @staticmethod
    def futures_last_price(raw_data: dict) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о фьючерсных ценах, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        return {item["symbol"]: float(item["lastPrice"]) for item in raw_data["data"]}

    @staticmethod
    def ticker_24hr(raw_data: list[dict]) -> TickerDailyDict:
        """Преобразует сырой ответ, в котором содержатся данные о статистике за 24 часа, в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.

        Возвращает:
            TickerDailyDict: Словарь, где ключ - тикер, а значение - статистика за последние 24 часа.
        """
        return {
            item["symbol"]: TickerDailyItem(
                p=round(float(item["priceChangePercent"]) * 100, 2),  # Конвертируем в проценты
                v=float(item["volume"]),
                q=float(item["quoteVolume"]),
            )
            for item in raw_data
        }

    @staticmethod
    def futures_ticker_24hr(raw_data: dict) -> TickerDailyDict:
        """Преобразует сырой ответ, в котором содержатся данные о фьючерсной статистике за 24 часа, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            TickerDailyDict: Словарь, где ключ - тикер, а значение - статистика за последние 24 часа.
        """
        result = {}
        for item in raw_data["data"]:
            symbol = item["symbol"]
            result[symbol] = TickerDailyItem(
                p=round(float(item["riseFallRate"]) * 100, 2),
                v=float(item["volume24"]) * Adapter._get_contract_size(symbol),
                q=float(item["amount24"]),
            )
        return result

    @staticmethod
    def open_interest(raw_data: dict) -> OpenInterestDict:
        """Преобразует сырой ответ, в котором содержатся данные об открытом интересе, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            OpenInterestDict: Словарь, где ключ - тикер, а значение - агрегированные данные открытого интереса.
        """
        result = {}
        for item in raw_data["data"]:
            symbol = item["symbol"]
            result[symbol] = OpenInterestItem(
                t=item["timestamp"],
                v=float(item["holdVol"]) * Adapter._get_contract_size(symbol),
                u="coins",
            )
        return result

    @staticmethod
    def funding_rate(raw_data: dict) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о ставках финансирования, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - ставка финансирования.
        """
        return {
            item["symbol"]: float(item["fundingRate"]) * 100
            for item in raw_data["data"]
            if "fundingRate" in item  # В некоторых элементах item нет ключа 'fundingRate'
        }

    @staticmethod
    def klines(raw_data: list[list], symbol: str) -> list[KlineDict]:
        """Преобразует сырой ответ, в котором содержатся данные о свечах, в унифицированный формат.

        Параметры:
            raw_data (list[list]): Сырой ответ с биржи.
            symbol (str): Символ тикера.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        return [
            KlineDict(
                s=symbol,
                t=kline[0],
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[7]),
                T=kline[6],
                x=None,
            )
            for kline in sorted(
                raw_data,
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def futures_klines(raw_data: dict, symbol: str) -> list[KlineDict]:
        """Преобразует сырой ответ, в котором содержатся данные о фьючерсных свечах, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.
            symbol (str): Символ тикера.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        data = raw_data["data"]

        times = data["time"]
        opens = data["open"]
        highs = data["high"]
        lows = data["low"]
        closes = data["close"]
        volumes = data["vol"]
        amounts = data["amount"]

        klines: list[KlineDict] = []

        for kline_time, open_, high, low, close, volume, amount in zip(
            times,
            opens,
            highs,
            lows,
            closes,
            volumes,
            amounts,
            strict=False,
        ):
            timestamp = int(float(kline_time))
            if timestamp < 10**12:
                timestamp *= 1000

            klines.append(
                KlineDict(
                    s=symbol,
                    t=timestamp,
                    o=float(open_),
                    h=float(high),
                    l=float(low),
                    c=float(close),
                    v=float(volume),
                    q=float(amount),
                    T=None,
                    x=None,
                )
            )

        return sorted(klines, key=lambda kline_item: kline_item["t"])

    @staticmethod
    def klines_message(raw_msg: Any) -> list[KlineDict]:
        """Преобразует вебсокет-сообщение со свечами в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список свечей в унифицированном формате.
        """
        kline = raw_msg["publicSpotKline"]
        return [
            KlineDict(
                s=raw_msg["symbol"],
                t=int(kline["windowStart"]) * 1000,
                o=float(kline["openingPrice"]),
                h=float(kline["highestPrice"]),
                l=float(kline["lowestPrice"]),
                c=float(kline["closingPrice"]),
                v=float(kline["volume"]),
                T=int(kline["windowEnd"]) * 1000,
                x=None,
                q=float(kline["amount"]),
            )
        ]

    @staticmethod
    def futures_klines_message(raw_msg: Any) -> list[KlineDict]:
        """Преобразует вебсокет-сообщение со свечами в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список свечей в унифицированном формате.
        """
        data = raw_msg["data"]
        return [
            KlineDict(
                s=data["symbol"],
                t=data["t"] * 1000,
                o=data["o"],
                h=data["h"],
                l=data["l"],
                c=data["c"],
                v=data["q"],  # Контракты
                q=data["a"],
                T=None,
                x=None,
            )
        ]

    @staticmethod
    def trades_message(raw_msg: Any) -> list[TradeDict]:
        """Преобразует вебсокет-сообщение со сделками в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[TradeDict]: Список сделок в унифицированном формате.
        """
        return [
            TradeDict(
                t=trade["time"],
                s=raw_msg["symbol"],
                S="BUY" if trade["tradeType"] == 1 else "SELL",
                p=float(trade["price"]),
                v=float(trade["quantity"]),
            )
            for trade in sorted(
                raw_msg["publicAggreDeals"]["deals"],
                key=lambda item: item["time"],
            )
        ]

    @staticmethod
    def futures_trades_message(raw_msg: Any) -> list[TradeDict]:
        """Преобразует вебсокет-сообщение со сделками в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[TradeDict]: Список сделок в унифицированном формате.
        """
        return [
            TradeDict(
                t=item["t"],
                s=raw_msg["symbol"],
                S="BUY" if item["T"] == 1 else "SELL",
                p=item["p"],
                v=item["v"] * Adapter._get_contract_size(raw_msg["symbol"]),
            )
            for item in sorted(
                raw_msg["data"],
                key=lambda item: item["t"],
            )
        ]

    @staticmethod
    def _get_contract_size(symbol: str) -> float:
        """Возвращает размер контракта для указанного символа тикера."""
        try:
            return ExchangeInfo.get_futures_ticker_info(symbol)["contract_size"] or 1
        except:  # noqa
            return 1



__all__ = ["UniClient"]


from typing import overload

from unicex._abc import IUniClient
from unicex.enums import Exchange, MarketType, Timeframe
from unicex.types import KlineDict, OpenInterestItem, TickerDailyDict

from .adapter import Adapter
from .client import Client


class UniClient(IUniClient[Client]):
    """Унифицированный клиент для работы с Binance API."""

    @property
    def _client_cls(self) -> type[Client]:
        """Возвращает класс клиента для Binance.

        Возвращает:
            type[Client]: Класс клиента для Binance.
        """
        return Client

    async def tickers(self, only_usdt: bool = True) -> list[str]:
        """Возвращает список тикеров.

        Параметры:
            only_usdt (bool): Если True, возвращает только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        raw_data = await self._client.ticker_price()
        return Adapter.tickers(raw_data=raw_data, only_usdt=only_usdt)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    async def futures_tickers(self, only_usdt: bool = True) -> list[str]:
        """Возвращает список тикеров.

        Параметры:
            only_usdt (bool): Если True, возвращает только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        raw_data = await self._client.futures_ticker_price()
        return Adapter.tickers(raw_data=raw_data, only_usdt=only_usdt)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    async def last_price(self) -> dict[str, float]:
        """Возвращает последнюю цену для каждого тикера.

        Возвращает:
            dict[str, float]: Словарь с последними ценами для каждого тикера.
        """
        raw_data = await self._client.ticker_price()
        return Adapter.last_price(raw_data)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    async def futures_last_price(self) -> dict[str, float]:
        """Возвращает последнюю цену для каждого тикера.

        Возвращает:
            dict[str, float]: Словарь с последними ценами для каждого тикера.
        """
        raw_data = await self._client.futures_ticker_price()
        return Adapter.last_price(raw_data)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    async def ticker_24hr(self) -> TickerDailyDict:
        """Возвращает статистику за последние 24 часа для каждого тикера.

        Возвращает:
            TickerDailyDict: Словарь с статистикой за последние 24 часа для каждого тикера.
        """
        raw_data = await self._client.ticker_24hr()
        return Adapter.ticker_24hr(raw_data=raw_data)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    async def futures_ticker_24hr(self) -> TickerDailyDict:
        """Возвращает статистику за последние 24 часа для каждого тикера.

        Возвращает:
            TickerDailyDict: Словарь с статистикой за последние 24 часа для каждого тикера.
        """
        raw_data = await self._client.futures_ticker_24hr()
        return Adapter.ticker_24hr(raw_data=raw_data)  # type: ignore | raw_data is list[dict] if symbol param is not ommited

    async def klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        """Возвращает список свечей для тикера.

        Параметры:
            symbol (str): Название тикера.
            limit (int | None): Количество свечей.
            interval (Timeframe | str): Таймфрейм свечей.
            start_time (int | None): Время начала периода в миллисекундах.
            end_time (int | None): Время окончания периода в миллисекундах.

        Возвращает:
            list[KlineDict]: Список свечей для тикера.
        """
        interval = (
            interval.to_exchange_format(Exchange.BINANCE, MarketType.SPOT)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.klines(
            symbol=symbol,
            interval=interval,
            limit=limit,
            start_time=start_time,
            end_time=end_time,
        )
        return Adapter.klines(raw_data=raw_data, symbol=symbol)

    async def futures_klines(
        self,
        symbol: str,
        interval: Timeframe | str,
        limit: int | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
    ) -> list[KlineDict]:
        """Возвращает список свечей для тикера.

        Параметры:
            symbol (str): Название тикера.
            limit (int | None): Количество свечей.
            interval (Timeframe | str): Таймфрейм свечей.
            start_time (int | None): Время начала периода в миллисекундах.
            end_time (int | None): Время окончания периода в миллисекундах.

        Возвращает:
            list[KlineDict]: Список свечей для тикера.
        """
        interval = (
            interval.to_exchange_format(Exchange.BINANCE, MarketType.FUTURES)
            if isinstance(interval, Timeframe)
            else interval
        )
        raw_data = await self._client.futures_klines(
            symbol=symbol,
            interval=interval,
            limit=limit,
            start_time=start_time,
            end_time=end_time,
        )
        return Adapter.klines(raw_data=raw_data, symbol=symbol)

    @overload
    async def funding_rate(self, symbol: str) -> float: ...

    @overload
    async def funding_rate(self, symbol: None) -> dict[str, float]: ...

    @overload
    async def funding_rate(self) -> dict[str, float]: ...

    async def funding_rate(self, symbol: str | None = None) -> dict[str, float] | float:
        """Возвращает ставку финансирования для тикера или всех тикеров, если тикер не указан.

        Параметры:
          symbol (`str | None`): Название тикера (Опционально).

        Возвращает:
          `dict[str, float] | float`: Ставка финансирования для тикера или словарь со ставками для всех тикеров.
        """
        raw_data = await self._client.futures_mark_price(symbol=symbol)
        adapted_data = Adapter.funding_rate(raw_data if isinstance(raw_data, list) else [raw_data])  # type: ignore[arg-type]
        return adapted_data[symbol] if symbol else adapted_data

    async def open_interest(self, symbol: str = None) -> OpenInterestItem:  # type: ignore[reportArgumentType] | We should provide our exception message
        """Возвращает объем открытого интереса для тикера.

        Параметры:
            symbol (`str`): Название тикера.

        Возвращает:
            `OpenInterestItem`: Словарь со временем и объемом открытого интереса в монетах.
        """
        if not symbol:
            raise ValueError("Symbol is required for binance open interest")
        raw_data = await self._client.open_interest(symbol=symbol)
        return Adapter.open_interest(raw_data)



__all__ = ["Adapter"]

from typing import Any

from unicex.types import (
    KlineDict,
    LiquidationDict,
    OpenInterestItem,
    TickerDailyDict,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с BingX API."""

    @staticmethod
    def tickers(raw_data: dict, only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о тикерах в список тикеров.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        data = raw_data["data"]
        items = data["symbols"] if isinstance(data, dict) and "symbols" in data else data
        return [
            item["symbol"] for item in items if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def ticker_24hr(raw_data: dict) -> TickerDailyDict:
        """Преобразует сырой ответ, в котором содержатся данные о тикере за последние 24 часа в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            TickerDailyDict: Словарь, где ключ - тикер, а значение - статистика за последние 24 часа.
        """
        items = raw_data["data"]
        result = {}
        for item in items:
            percent_raw = item["priceChangePercent"]
            percent_str = str(percent_raw).strip()
            if percent_str.endswith("%"):
                percent_str = percent_str[:-1]
            result[item["symbol"]] = {
                "p": float(percent_str),
                "v": float(item["volume"]),
                "q": float(item["quoteVolume"]),
            }
        return result

    @staticmethod
    def open_interest(raw_data: dict) -> OpenInterestItem:
        """Преобразует сырой ответ, в котором содержатся данные об открытом интересе, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            OpenInterestItem: Словарь со временем и объемом открытого интереса в монетах.
        """
        item = raw_data["data"]
        return OpenInterestItem(t=int(item["time"]), v=float(item["openInterest"]), u="usd")

    @staticmethod
    def funding_rate(raw_data: dict) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о ставках финансирования, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - ставка финансирования.
        """
        data = raw_data["data"]
        items = data if isinstance(data, list) else [data]
        result: dict[str, float] = {}
        for item in items:
            if "lastFundingRate" in item:
                result[item["symbol"]] = float(item["lastFundingRate"]) * 100
            elif "fundingRate" in item and item["fundingRate"] != "":
                result[item["symbol"]] = float(item["fundingRate"]) * 100
        return result

    @staticmethod
    def last_price(raw_data: dict) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о последней цене, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        data = raw_data["data"]
        if isinstance(data, list):
            return {
                item["symbol"]: float(item["lastPrice"] if "lastPrice" in item else item["price"])
                for item in data
            }
        if isinstance(data, dict):
            return {data["symbol"]: float(data["price"])}
        return {}

    @staticmethod
    def klines(raw_data: dict) -> list[KlineDict]:
        """Преобразует сырой ответ, в котором содержатся данные о свечах, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        data = raw_data["data"]
        items = data["klines"] if isinstance(data, dict) and "klines" in data else data
        if "symbol" in raw_data:
            symbol = raw_data["symbol"]
        elif isinstance(data, dict) and "symbol" in data:
            symbol = data["symbol"]
        else:
            symbol = ""

        if items and isinstance(items[0], dict):
            return [
                KlineDict(
                    s=symbol,
                    t=int(kline["time"]),
                    o=float(kline["open"]),
                    h=float(kline["high"]),
                    l=float(kline["low"]),
                    c=float(kline["close"]),
                    v=float(kline["volume"]),
                    q=0.0,
                    T=None,
                    x=None,
                )
                for kline in sorted(
                    items,
                    key=lambda x: int(x["time"]),
                )
            ]

        return [
            KlineDict(
                s=symbol,
                t=int(kline[0]),
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[7]) if len(kline) > 7 else 0.0,
                T=int(kline[6]) if len(kline) > 6 else None,
                x=None,
            )
            for kline in sorted(
                items,
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def Klines_message(msg: Any) -> list[KlineDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        свече/свечах в унифицированный вид.

        Параметры:
            msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        ...

    @staticmethod
    def futures_klines_message(msg: Any) -> list[KlineDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        свече/свечах в унифицированный вид.

        Параметры:
            msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        ...

    @staticmethod
    def aggtrades_message(msg: Any) -> list[TradeDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        аггрегированных сделке/сделках в унифицированный вид.

        Параметры:
            msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о сделке.
        """
        ...

    @staticmethod
    def futures_aggtrades_message(msg: Any) -> list[TradeDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        аггрегированных сделке/сделках в унифицированный вид.

        Параметры:
            msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о сделке.
        """
        ...

    @staticmethod
    def trades_message(msg: Any) -> list[TradeDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        сделке/сделках в унифицированный вид.

        Параметры:
            msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о сделке.
        """
        return [
            TradeDict(
                t=int(trade["T"]),
                s=str(trade["s"]),
                S="SELL" if bool(trade["m"]) else "BUY",
                p=float(trade["p"]),
                v=float(trade["q"]),
            )
            for trade in sorted(
                msg["data"],
                key=lambda x: int(x["T"]),
            )
        ]

    @staticmethod
    def futures_trades_message(msg: Any) -> list[TradeDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        сделке/сделках в унифицированный вид.

        Параметры:
            msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о сделке.
        """
        return [
            TradeDict(
                t=int(trade["T"]),
                s=str(trade["s"]),
                S="SELL" if bool(trade["m"]) else "BUY",
                p=float(trade["p"]),
                v=float(trade["q"]),
            )
            for trade in sorted(
                msg["data"],
                key=lambda x: int(x["T"]),
            )
        ]

    @staticmethod
    def liquidations_message(msg: Any) -> list[LiquidationDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        ликвидации/ликвидациях в унифицированный вид.

        Параметры:
            msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[LiquidationDict]: Список словарей, где каждый словарь содержит данные о ликвидации.
        """
        ...



from __future__ import annotations

import time
from typing import Any

__all__ = ["Adapter"]

from unicex.types import (
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods

from .exchange_info import ExchangeInfo


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Gateio API."""

    @staticmethod
    def tickers(raw_data: list[dict], only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ о тикерах в список символов.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре c USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["currency_pair"]
            for item in raw_data
            if item["currency_pair"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def futures_tickers(raw_data: list[dict], only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ о фьючерсных тикерах в список символов.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре c USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["contract"]
            for item in raw_data
            if item["contract"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def last_price(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует данные о последних ценах (spot) в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ — тикер, а значение — последняя цена.
        """
        return {item["currency_pair"]: float(item["last"]) for item in raw_data}

    @staticmethod
    def futures_last_price(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует данные о последних ценах (futures) в унифицированный формат."""
        return {item["contract"]: float(item["last"]) for item in raw_data}

    @staticmethod
    def ticker_24hr(raw_data: list[dict]) -> TickerDailyDict:
        """Преобразует 24-часовую статистику (spot) в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.

        Возвращает:
            TickerDailyDict: Словарь, где ключ — тикер, а значение — агрегированная статистика.
        """
        return {
            item["currency_pair"]: TickerDailyItem(
                p=float(item["change_percentage"]),
                v=float(item["base_volume"]),
                q=float(item["quote_volume"]),
            )
            for item in raw_data
        }

    @staticmethod
    def futures_ticker_24hr(raw_data: list[dict]) -> TickerDailyDict:
        """Преобразует 24-часовую статистику (futures) в унифицированный формат."""
        return {
            item["contract"]: TickerDailyItem(
                p=float(item["change_percentage"]),
                v=float(item["volume_24h_base"]),
                q=float(item["volume_24h_quote"]),
            )
            for item in raw_data
        }

    @staticmethod
    def klines(raw_data: list[list], symbol: str) -> list[KlineDict]:
        """Преобразует данные о свечах в унифицированный формат.

        Параметры:
            raw_data (list[list]): Сырой ответ с биржи.
            symbol (str): Символ тикера.

        Возвращает:
            list[KlineDict]: Список свечей.
        """
        return [
            KlineDict(
                s=symbol,
                t=int(kline[0]) * 1000,  # переводим секунды → миллисекунды
                o=float(kline[5]),
                h=float(kline[3]),
                l=float(kline[4]),
                c=float(kline[2]),
                v=float(kline[6]),
                q=float(kline[1]),
                T=None,
                x=kline[7] == "true",
            )
            for kline in sorted(
                raw_data,
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def futures_klines(raw_data: list[dict], symbol: str) -> list[KlineDict]:
        """Преобразует данные о свечах в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.
            symbol (str): Символ тикера.

        Возвращает:
            list[KlineDict]: Список свечей.
        """
        return [
            KlineDict(
                s=symbol,
                t=int(kline["t"]) * 1000,  # переводим секунды → миллисекунды
                o=float(kline["o"]),
                h=float(kline["h"]),
                l=float(kline["l"]),
                c=float(kline["c"]),
                v=float(kline["v"]),
                q=float(kline["sum"]),  # "sum" = объем в $ (quote volume)
                T=None,
                x=None,
            )
            for kline in sorted(raw_data, key=lambda x: int(x["t"]))
        ]

    @staticmethod
    def funding_rate(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует данные о ставках финансирования в унифицированный формат."""
        return {
            item["contract"]: float(item["funding_rate"]) * 100
            for item in raw_data
            if item.get("funding_rate") is not None
        }

    @staticmethod
    def open_interest(raw_data: list[dict]) -> OpenInterestDict:
        """Преобразует данные об открытом интересе в унифицированный формат."""
        return {
            item["contract"]: OpenInterestItem(
                t=int(time.time() * 1000),
                v=float(item["total_size"]) * float(item["quanto_multiplier"]),
                u="coins",
            )
            for item in raw_data
        }

    @staticmethod
    def klines_message(raw_msg: Any) -> list[KlineDict]:
        """Преобразует вебсокет-сообщение со свечами в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список свечей в унифицированном формате.
        """
        data = raw_msg["result"]
        return [
            KlineDict(
                s=data["n"].split("_", 1)[1],  # XRP_USDT
                t=int(data["t"]) * 1000,
                o=float(data["o"]),
                h=float(data["h"]),
                l=float(data["l"]),
                c=float(data["c"]),
                v=float(data["a"]),
                q=float(data["v"]),
                T=None,
                x=not data["w"],  # w=False → свеча закрыта
            )
        ]

    @staticmethod
    def futures_klines_message(raw_msg: Any) -> list[KlineDict]:
        """Преобразует вебсокет-сообщение со свечами в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список свечей в унифицированном формате.
        """
        return [
            KlineDict(
                s=item["n"].split("_", 1)[1],  # XRP_USDT
                t=int(item["t"]) * 1000,
                o=float(item["o"]),
                h=float(item["h"]),
                l=float(item["l"]),
                c=float(item["c"]),
                v=float(item["a"]),
                q=float(item["v"]),
                T=None,
                x=not item["w"],  # w=False → свеча закрыта
            )
            for item in sorted(
                raw_msg["result"],
                key=lambda x: int(x["t"]),
            )
        ]

    @staticmethod
    def trades_message(raw_msg: Any) -> list[TradeDict]:
        """Преобразует вебсокет-сообщение со сделками в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[TradeDict]: Список сделок в унифицированном формате.
        """
        trade = raw_msg["result"]
        return [
            TradeDict(
                t=trade["create_time_ms"],
                s=trade["currency_pair"],
                S=trade["side"].upper(),
                p=float(trade["price"]),
                v=float(trade["amount"]),
            )
        ]

    @staticmethod
    def futures_trades_message(raw_msg: Any) -> list[TradeDict]:
        """Преобразует вебсокет-сообщение со сделками в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[TradeDict]: Список сделок в унифицированном формате.
        """
        return [
            TradeDict(
                t=item["create_time_ms"],
                s=item["contract"],
                S="BUY" if float(item["size"]) > 0 else "SELL",
                p=float(item["price"]),
                v=abs(float(item["size"])) * Adapter._get_contract_size(item["contract"]),
            )
            for item in sorted(
                raw_msg["result"],
                key=lambda x: x["create_time_ms"],
            )
        ]

    @staticmethod
    def _get_contract_size(symbol: str) -> float:
        """Возвращает размер контракта для указанного символа тикера."""
        try:
            return ExchangeInfo.get_futures_ticker_info(symbol)["contract_size"] or 1
        except:  # noqa
            return 1



__all__ = ["WebsocketManager"]

import json
from collections.abc import Awaitable, Callable
from typing import Any

from unicex._base import Websocket

from .client import Client

type CallbackType = Callable[[Any], Awaitable[None]]


class WebsocketManager:
    """Менеджер асинхронных вебсокетов для Hyperliquid."""

    _URL: str = "wss://api.hyperliquid.xyz/ws"
    """Базовый URL для вебсокета."""

    def __init__(self, client: Client | None = None, **ws_kwargs: Any) -> None:
        """Инициализирует менеджер вебсокетов для Hyperliquid.

        Параметры:
            client (`Client | None`): Клиент для выполнения запросов. Нужен, чтобы открыть приватные вебсокеты.
            ws_kwargs (`dict[str, Any]`): Дополнительные аргументы, которые прокидываются в `Websocket`.
        """
        self.client = client
        self._ws_kwargs = {
            "ping_message": json.dumps({"method": "ping"}),
            "ping_interval": 30,
            **ws_kwargs,
        }

    def _create_subscription_message(self, subscription_type: str, **params: Any) -> str:
        """Создает сообщение подписки для Hyperliquid.

        Параметры:
            subscription_type (`str`): Тип подписки.
            params (`dict[str, Any]`): Параметры подписки.

        Возвращает:
            `str`: JSON-строка с сообщением подписки.
        """
        subscription = {"type": subscription_type, **params}
        message = {"method": "subscribe", "subscription": subscription}
        return json.dumps(message)

    def all_mids(self, callback: CallbackType, dex: str | None = None) -> Websocket:
        """Создает вебсокет для получения средних цен всех активов.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            dex (`str | None`): Perp dex для получения средних цен. Если не указан, используется первый perp dex.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        params = {}
        if dex is not None:
            params["dex"] = dex

        subscription_message = self._create_subscription_message("allMids", **params)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def notification(self, callback: CallbackType, user: str) -> Websocket:
        """Создает вебсокет для получения уведомлений пользователя.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message("notification", user=user)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def web_data2(self, callback: CallbackType, user: str) -> Websocket:
        """Создает вебсокет для получения веб-данных пользователя.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message("webData2", user=user)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def candle(
        self,
        callback: CallbackType,
        interval: str,
        coin: str | None = None,
        coins: list[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения свечей.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            coin (`str`): Символ монеты.
            coins (`list[str]`): Список символов монет для мультиплекс подключения.
            interval (`str`): Интервал свечей ("1m", "3m", "5m", "15m", "30m", "1h", "2h", "4h", "8h", "12h", "1d", "3d", "1w", "1M").

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        if coin and coins:
            raise ValueError("Parameters coin and coins cannot be used together")
        if not (coin or coins):
            raise ValueError("Either coin or coins must be provided")
        if coin:
            coins = [coin]
        subscription_messages = [
            self._create_subscription_message("candle", coin=coin, interval=interval)
            for coin in coins  # type: ignore
        ]
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=subscription_messages,
            **self._ws_kwargs,
        )

    def l2_book(
        self,
        callback: CallbackType,
        coin: str,
        n_sig_figs: int | None = None,
        mantissa: int | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения стакана L2.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            coin (`str`): Символ монеты.
            n_sig_figs (`int | None`): Количество значащих цифр для округления цен.
            mantissa (`int | None`): Мантисса для округления цен.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        params = {"coin": coin}
        if n_sig_figs is not None:
            params["nSigFigs"] = n_sig_figs  # type: ignore
        if mantissa is not None:
            params["mantissa"] = mantissa  # type: ignore

        subscription_message = self._create_subscription_message("l2Book", **params)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def trades(
        self, callback: CallbackType, coin: str | None = None, coins: list[str] | None = None
    ) -> Websocket:
        """Создает вебсокет для получения сделок.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            coin (`str | None`): Символ монеты.
            coins (`list[str] | None`): Список символов монет для мультиплекс подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        if coin and coins:
            raise ValueError("Parameters coin and coins cannot be used together")
        if not (coin or coins):
            raise ValueError("Either coin or coins must be provided")
        if coin:
            coins = [coin]
        subscription_messages = [
            self._create_subscription_message("trades", coin=coin)
            for coin in coins  # type: ignore
        ]
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=subscription_messages,
            **self._ws_kwargs,
        )

    def order_updates(self, callback: CallbackType, user: str) -> Websocket:
        """Создает вебсокет для получения обновлений ордеров пользователя.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message("orderUpdates", user=user)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def user_events(self, callback: CallbackType, user: str) -> Websocket:
        """Создает вебсокет для получения событий пользователя.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message("userEvents", user=user)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def user_fills(
        self, callback: CallbackType, user: str, aggregate_by_time: bool | None = None
    ) -> Websocket:
        """Создает вебсокет для получения исполнений пользователя.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.
            aggregate_by_time (`bool | None`): Агрегировать по времени.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        params = {"user": user}
        if aggregate_by_time is not None:
            params["aggregateByTime"] = aggregate_by_time  # type: ignore

        subscription_message = self._create_subscription_message("userFills", **params)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def user_fundings(self, callback: CallbackType, user: str) -> Websocket:
        """Создает вебсокет для получения фандинга пользователя.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message("userFundings", user=user)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def user_non_funding_ledger_updates(self, callback: CallbackType, user: str) -> Websocket:
        """Создает вебсокет для получения обновлений леджера пользователя (кроме фандинга).

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message(
            "userNonFundingLedgerUpdates", user=user
        )
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def active_asset_ctx(self, callback: CallbackType, coin: str) -> Websocket:
        """Создает вебсокет для получения контекста активного актива.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            coin (`str`): Символ монеты.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message("activeAssetCtx", coin=coin)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def active_asset_data(self, callback: CallbackType, user: str, coin: str) -> Websocket:
        """Создает вебсокет для получения данных активного актива пользователя (только Perps).

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.
            coin (`str`): Символ монеты.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message(
            "activeAssetData", user=user, coin=coin
        )
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def user_twap_slice_fills(self, callback: CallbackType, user: str) -> Websocket:
        """Создает вебсокет для получения TWAP slice fills пользователя.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message("userTwapSliceFills", user=user)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def user_twap_history(self, callback: CallbackType, user: str) -> Websocket:
        """Создает вебсокет для получения TWAP истории пользователя.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            user (`str`): Адрес пользователя.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message("userTwapHistory", user=user)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )

    def bbo(self, callback: CallbackType, coin: str) -> Websocket:
        """Создает вебсокет для получения лучшего бида/аска.

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            coin (`str`): Символ монеты.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_message = self._create_subscription_message("bbo", coin=coin)
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=[subscription_message],
            **self._ws_kwargs,
        )



__all__ = ["Adapter"]

from typing import Any

from unicex.types import (
    KlineDict,
    LiquidationDict,
    OpenInterestDict,
    OpenInterestItem,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Bybit API."""

    @staticmethod
    def tickers(raw_data: dict, only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о тикерах в список тикеров.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["symbol"]
            for item in raw_data["result"]["list"]
            if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def ticker_24hr(raw_data: dict) -> TickerDailyDict:
        """Преобразует сырой ответ, в котором содержатся данные о тикере за последние 24 часа в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            TickerDailyDict: Словарь, где ключ - тикер, а значение - статистика за последние 24 часа.
        """
        return {
            item["symbol"]: TickerDailyItem(
                p=round(float(item["price24hPcnt"]) * 100, 2),
                v=float(item["volume24h"]),
                q=float(item["turnover24h"]),
            )
            for item in raw_data["result"]["list"]
        }

    @staticmethod
    def open_interest(raw_data: dict) -> OpenInterestDict:
        """Преобразует сырой ответ, в котором содержатся данные об открытом интересе, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            OpenInterestDict: Словарь, где ключ - тикер, а значение - агрегированные данные открытого интереса.
        """
        return {
            item["symbol"]: OpenInterestItem(
                t=raw_data["time"],
                v=float(item["openInterest"]),
                u="coins",
            )
            for item in raw_data["result"]["list"]
        }

    @staticmethod
    def funding_rate(raw_data: dict) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о ставках финансирования, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - ставка финансирования.
        """
        return {
            item["symbol"]: float(item["fundingRate"]) * 100
            for item in raw_data["result"]["list"]
            if item["fundingRate"]
        }

    @staticmethod
    def last_price(raw_data: dict) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о последней цене, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        return {item["symbol"]: float(item["lastPrice"]) for item in raw_data["result"]["list"]}

    @staticmethod
    def klines(raw_data: dict) -> list[KlineDict]:
        """Преобразует сырой ответ, в котором содержатся данные о свечах, в унифицированный формат.

        Параметры:
            raw_data (dict): Сырой ответ с биржи.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        return [
            KlineDict(
                s=raw_data["result"]["symbol"],
                t=int(kline[0]),
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[6]),
                T=None,
                x=None,
            )
            for kline in sorted(
                raw_data["result"]["list"],
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def Klines_message(raw_msg: Any) -> list[KlineDict]:
        """Преобразует вебсокет-сообщение с данными о свечах в унифицированный формат.

        Параметры:
        raw_msg (`Any`): Сырое сообщение из вебсокета Bybit.

        Возвращает:
          `list[KlineDict]`: Список свечей в унифицированном формате.
        """
        symbol = raw_msg["topic"].split(".")[-1]
        return [
            KlineDict(
                s=symbol,
                t=kline["start"],
                o=float(kline["open"]),
                h=float(kline["high"]),
                l=float(kline["low"]),
                c=float(kline["close"]),
                v=float(kline["volume"]),
                q=float(kline["turnover"]),
                T=kline["end"],
                x=kline["confirm"],
            )
            for kline in sorted(
                raw_msg["data"],
                key=lambda x: int(x["start"]),
            )
        ]

    @staticmethod
    def trades_message(raw_msg: Any) -> list[TradeDict]:
        """Преобразует вебсокет-сообщение с данными о сделках в унифицированный формат.

        Параметры:
        raw_msg (`Any`): Сырое сообщение из вебсокета Bybit.

        Возвращает:
          `list[TradeDict]`: Список сделок в унифицированном формате.
        """
        return [
            TradeDict(
                t=trade["T"],
                s=trade["s"],
                S=trade["S"].upper(),
                p=float(trade["p"]),
                v=float(trade["v"]),
            )
            for trade in sorted(
                raw_msg["data"],
                key=lambda x: int(x["T"]),
            )
        ]

    @staticmethod
    def liquidations_message(raw_msg: Any) -> list[LiquidationDict]:
        """Преобразует вебсокет-сообщение с данными о ликвидациях в унифицированный формат.

        Параметры:
        raw_msg (`Any`): Сырое сообщение из вебсокета Bybit.

        Возвращает:
          `list[LiquidationDict]`: Список ликвидаций в унифицированном формате.
        """
        return [
            LiquidationDict(
                t=liquidation["T"],
                s=liquidation["s"],
                S=liquidation["S"].upper(),
                v=float(liquidation["v"]),
                p=float(liquidation["p"]),
            )
            for liquidation in sorted(
                raw_msg["data"],
                key=lambda x: int(x["T"]),
            )
        ]



__all__ = ["ExchangeInfo"]

import aiohttp

from unicex._abc import IExchangeInfo


class ExchangeInfo(IExchangeInfo):
    """Предзагружает информацию о тикерах для биржи BingX."""

    exchange_name = "BingX"
    """Название биржи, на которой работает класс."""

    @classmethod
    async def _load_spot_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для спотового рынка."""
        ...

    @classmethod
    async def _load_futures_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для фьючерсного рынка."""
        ...



"""Модуль,который описывает исключения и ошибки, которые могут возникнуть при работе с библиотекой."""

from dataclasses import dataclass, field


@dataclass
class UniCexException(Exception):
    """Базовое исключение библиотеки."""

    message: str
    """Сообщение об ошибке."""


@dataclass
class NotAuthorized(UniCexException):
    """Исключение, возникающее при отсутствии авторизации."""

    pass


@dataclass
class NotSupported(UniCexException):
    """Исключение, возникающее при попытке использования не поддерживаемой функции."""

    pass


@dataclass
class AdapterError(UniCexException):
    """Исключение, возникающее при ошибке адаптации данных."""

    pass


@dataclass
class QueueOverflowError(UniCexException):
    """Исключение, возникающее при переполнении очереди сообщений."""

    pass


@dataclass
class ResponseError(UniCexException):
    """Исключение, возникающее при ошибке ответа."""

    status_code: int
    code: str = ""  # "" - means undefined
    response_json: dict = field(default_factory=dict)
    response_text: str = ""

    def __str__(self) -> str:
        """Возвращает строковое представление исключения."""
        if self.response_json:
            preview = str(self.response_json)
            if len(preview) > 500:
                preview = preview[:500] + "..."
            return f"ResponseError: status_code={self.status_code}, code={self.code}, response_json: {preview}"
        elif self.response_text:
            preview = str(self.response_text)
            if len(preview) > 500:
                preview = preview[:500] + "..."
            return f"ResponseError: status_code={self.status_code}, code={self.code}, response_text: {preview}"
        else:
            return f"ResponseError: status_code={self.status_code}, code={self.code}"
